# -*- coding: UTF-8 -*-
"""Module for defining report element classes.

"""
from .base import *
from ..helpers.data.transform import json2html, json2xml


__all__ = __features__ = ["Code", "Data", "Footer", "Header", "List", "Section", "Subsection", "Table", "Text", "Title"]

HEAD_CSS = "@%(pos)s-left{%(left)s};@%(pos)s-center{%(center)s};@%(pos)s-right{%(right)s};"


class Data(Element):
    """ This class represents a data dictionary.
    
    :param data: data dictionary
    """
    filename = "data"
    
    def __init__(self, data, **kwargs):
        super(Data, self).__init__(**kwargs)
        if not isinstance(data, dict):
            raise ValueError("'data' argument shall be a dictionary")
        self.data = data
    
    @output
    def html(self, indent=4, text=TEXT):
        """ Generate an HTML table from the data dictionary. """
        return json2html(self.data)
    
    @output
    def json(self, text=TEXT):
        """ Return the original JSON object. """
        return self.data
    
    @output
    def xml(self, indent=2, text=TEXT):
        """ Generate an XML output from the data dictionary. """
        return json2xml(self.data)


class Footer(Element):
    """ This class represents the footer of a report.
    
    :param left:   left-side section of the footer
    :param center: center section of the footer
    :param right:  right-side section of the footer
    """
    _style = {'size': 9, 'style': "normal", 'color': "black"}
    pos = "bottom"
    
    def __init__(self, left="", center="", right="", **kwargs):
        super(Footer, self).__init__(**kwargs)
        self.pos = self.__class__.pos
        if self.pos == "bottom" and center == "":
            center = "\" counter(page) \"/\" counter(pages) \""
        self.data = {'left': left, 'center': center, 'right': right}
        d = {k: v for k, v in self.data.items()}
        d.update(pos=self.pos)
        self.css = HEAD_CSS % d
    
    @output
    def xml(self, indent=2, text=TEXT):
        ind, nl = self._set_indent(indent)
        r = []
        for attr in ["left", "center", "right"]:
            if len(self.data[attr]) > 0:
                r.append(ind + "<%(tag)s>%(content)s</%(tag)s>" % {'tag': attr, 'content': self.data[attr]})
        if len(r) > 0:
            r = ["<%s>" % self.name] + r + ["</%s>" % self.name]
        return nl.join(r)


class Header(Footer):
    """ This class represents the header of a report.
    
    :param left:   left-side section of the header
    :param center: center section of the header
    :param right:  right-side section of the header
    """
    pos = "top"


class List(Element):
    """ This class represents a list of items, ordered or not.
    
    :param items:   list's items
    :param ordered: whether the list is to be ordered or not
    """
    def __init__(self, *items, **kwargs):
        super(List, self).__init__(**kwargs)
        self.data = items
        self.ordered = kwargs.get('ordered', False)
        self.tag = ["ul", "ol"][self.ordered]
    
    @output
    def csv(self, text=TEXT):
        """ Generate a dummy CSV table from the list. """
        return "\n".join(self.data)
    
    @output
    def html(self, indent=4, text=TEXT):
        ind, nl = self._set_indent(indent)
        s = '<%(tag)s style="%(style)s">' % self.__dict__ + nl
        for i in self.data:
            s += ind + "<li>%s</li>" % i + nl
        return s + "</%s>" % self.tag
    
    @output
    def md(self, text=TEXT):
        """ Generate Markdown from the list. """
        return "\n".join("%s %s" % (["-", "%d." % (n+1)][self.ordered], i) for n, i in enumerate(self.data))
    
    @output
    def xml(self, indent=2, text=TEXT):
        """ Generate an XML output from the list. """
        ind, nl = self._set_indent(indent)
        xml = "<%(name)s>{0}{1}{0}</%(name)s>" % self.__dict__
        return xml.format(nl, "\n".join(ind + "<item>%s</item>" % i for i in self.data))


class Table(Element):
    """ This class represents a table.
    
    :param data:           table data
    :param column_headers: list of column headers or "indices" or None ; if "indices", it is replaced accordingly
    :param row_headers:    list of row headers or "indices" or None ; if "indices", it is replaced accordingly
    :param column_footers: list of column footers or None
    :param flt_fmt:        float format to be displayed when generating the output
    """
    filename = "table"
    
    def __init__(self, data, column_headers="indices", row_headers=None, column_footers=None, flt_fmt="%.2g", **kwargs):
        super(Table, self).__init__(**kwargs)
        self.index = row_headers == "indices"
        self.column_headers = list(map(str, range(len(data[0])))) if column_headers == "indices" else column_headers
        self.column_footers = column_footers
        self.row_headers = list(map(str, range(len(data)))) if row_headers == "indices" else row_headers
        if self.row_headers and len(data) != len(self.row_headers):
            raise ValueError("Bad row headers length")
        if self.column_headers and len(data[0]) != len(self.column_headers):
            raise ValueError("Bad column headers length")
        if self.column_headers is not None and self.row_headers is not None:
            self.column_headers = [""] + self.column_headers
        self.data = data
        self.float_fmt = flt_fmt
    
    def _format(self):
        data = []
        for i, row in enumerate(self.data):
            row = list(map(lambda x: self.float_fmt % x if isinstance(x, float) else str(x), row))
            if self.row_headers is not None:
                row.insert(0, self.row_headers[i])
            data.append(row)
        return data
    
    @output
    def csv(self, text=TEXT, sep=','):
        """ Generate a CSV table from the table data. """
        r = "" if self.column_headers is None else sep.join(self.column_headers)
        for row in self._format():
            r += "\n" + sep.join(row)
        return r
    
    @output
    def html(self, indent=4, text=TEXT):
        """ Generate an HTML table from the table data. """
        ind, nl = self._set_indent(indent)
        r = ['<table id="%(name)s" style="%(style)s">' % self.__dict__]
        if self.column_headers is not None:
            r.append(ind + "<thead>")
            for h in self.column_headers:
                r.append(2 * ind + "<th>%s</th>" % h)
            r.append(ind + "</thead>")
        r.append(ind + "<tbody>")
        for row in self._format():
            r.append(ind + "<tr>")
            for i, v in enumerate(row):
                xml = "<{0}>%s</{0}>".format(["td", "th"][i == 0 and self.row_headers is not None])
                r.append(2 * ind + xml % v)
            r.append(ind + "</tr>")
        r.append(ind + "</tbody>")
        if self.column_footers is not None:
            r.append(ind + "<tfoot>")
            for h in self.column_footers:
                r.append(2 * ind + "<th>%s</th>" % h)
            r.append(ind + "</tfoot>")
        r.append("</table>")
        return nl.join(r)
    
    @output
    def md(self, float_format="%.2g", text=TEXT):
        """ Generate Markdown from the table data. """
        r = [" | ".join(self.column_headers or list(map(str, range(len(self.data[0])))))]
        r.append(" | ".join("---" for i in range(len(self.data[0]))))
        for row in self._format():
            r.append(" | ".join(row))
        if self.column_footers is not None:
            r.append(" | ".join(map(lambda x: "**%s**" % x, self.column_footers)))
        return "\n".join(r)
    
    @output
    def xml(self, indent=2, text=TEXT):
        """ Generate an XML output from the report data. """
        ind, nl = self._set_indent(indent)
        r = ["<%s>" % self.name]
        for row in self.data:
            r.append(ind + "<row>")
            for i, v in enumerate(row):
                if self.column_headers is None:
                    n = ""
                elif i == 0 and self.index:
                    n = ' name="index"'
                else:
                    n = ' name="%s"' % self.column_headers[i]
                r.append(2 * ind + "<field%s>%s</field>" % (n, v))
            r.append(ind + "</row>")
        if self.column_footers is not None:
            r.append(ind + "<footer>")
            for h in self.column_footers:
                r.append(2 * ind + "<field>%s</field>" % h)
            r.append(ind + "</footer>")
        r.append("</%s>" % self.name)
        return nl.join(r)


class Text(Element):
    """ Text area report element.
    
    :param content: text content
    :param tag:     HTML tag to be used
    """
    def __init__(self, content, tag="p", **kwargs):
        super(Text, self).__init__(**kwargs)
        self.data = content
        self.tag = tag
    @output
    def html(self, indent=4, text=TEXT):
        return ('<%(tag)s style="%(style)s">%(data)s</%(tag)s>' % self.__dict__).replace("\n", "<br>")
    
    @output
    def md(self, text=TEXT):
        return self.data


class Code(Text):
    """ Code block report element.
    
    :param code:     code content
    :param language: code's language
    param hl_lines:  lines to be highlighted
    """
    _style = {'size': 10, 'style': "normal", 'color': "grey"}
    
    def __init__(self, code, language=None, hl_lines=None, **kwargs):
        super(Code, self).__init__(code, "pre", **kwargs)
        self.language = language
        self.hl_lines = hl_lines
    
    @output
    def html(self, indent=4, text=TEXT):
        s = "<pre"
        if self.language:
            s += ' class="%s hljs"' % self.language
        s += ' style="%s">' % self.style
        return s + str(self.data).replace("\n", "<br>") + "</pre>"
    
    @output
    def md(self, text=TEXT):
        s = "```"
        if self.language:
            s += self.language
        if self.hl_lines:
            s += " hl_lines=\"%s\"" % self.hl_lines
        return s + "\n%s\n```" % self.data


class Title(Text):
    """ Title report element.
    
    :param title: title content
    :param tag:   HTML tag to be used
    """
    def __init__(self, title, tag="h1", **kwargs):
        if tag not in ["h1", "h2", "h3", "h4", "h5", "h6"]:
            raise ValueError("Title tag should be \"h[1-6]\", not \"{}\"".format(tag))
        super(Title, self).__init__(title, tag, **kwargs)
    
    @output
    def md(self, text=TEXT):
        return "%(prefix)s %(content)s" % {'prefix': "#" * int(self.tag[-1]), 'content': self.data}


class Section(Title):
    """ Section report element.
    
    :param title: title content
    :param tag:   HTML tag to be used
    """
    def __init__(self, title, tag="h2", **kwargs):
        super(Section, self).__init__(title, tag, **kwargs)


class Subsection(Title):
    """ Subsection report element.
    
    :param title: title content
    :param tag:   HTML tag to be used
    """
    def __init__(self, title, tag="h3", **kwargs):
        super(Subsection, self).__init__(title, tag, **kwargs)




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Deprecation warnings, for use when no backward-compatibility provided.

"""
__all__ = __features__ = []


def __deprecated(old, new=None):
    msg = "'{}' has been deprecated".format(old)
    if new is not None:
        msg += ", please use '{}'".format(new)
    raise DeprecationWarning(msg)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Initialization of Tinyscript package.

"""

from .__info__ import __author__, __copyright__, __license__, __version__

from .handlers import *
from .helpers import *
from .loglib import *
from .parser import *
from .preimports import *
from .warnings import *

from .handlers import __features__ as _handlers
from .helpers import __features__ as _helpers
from .loglib import __features__ as _loglib
from .parser import __features__ as _parser
from .preimports import __features__ as _preimports
from .warnings import __features__ as _warnings


ts.__author__    = __author__
ts.__copyright__ = __copyright__
ts.__license__   = __license__
ts.__version__   = __version__


__all__ = _handlers + _helpers + _loglib + _parser + _preimports + _warnings



# -*- coding: UTF-8 -*-
"""Subpackage containing data-related helper functions.

"""
from .transform import *
from .types import *
from .utils import *

from .transform import __features__ as _transform
from .types import __features__ as _types
from .utils import __features__ as _utils


__all__ = __features__ = _transform + _types + _utils



# -*- coding: UTF-8 -*-
"""Common Python2/3 compatibility functions.

"""
from six import binary_type, string_types, text_type, u

from .constants import PYTHON3


__all__ = __features__ = ["b", "binary_type", "byteindex", "execfile", "ensure_binary", "ensure_str", "iterbytes",
                          "string_types", "text_type", "u"]


# see: http://python3porting.com/problems.html
byteindex = lambda d, i=None: d[i] if PYTHON3 else ord(d[i])


def b(s):
    """
    Similar to six.b function, because the behavior of 'b' in Python2/3 is not exactly the same. This makes 'b' behave
     in Python 3 like in Python 2.
    """
    if PYTHON3:
        try:
            return s.encode("latin-1")
        except:
            pass
        try:
            return s.encode("utf-8")
        except:
            pass
    return s


def ensure_binary(s, encoding='utf-8', errors='strict'):
    """
    Identical to six.ensure_binary. Copied here to avoid messing up with six version errors.
    """
    if isinstance(s, text_type):
        return s.encode(encoding, errors)
    elif isinstance(s, binary_type):
        return s
    else:
        raise TypeError("not expecting type '%s'" % type(s))


def ensure_str(s, encoding='utf-8', errors='strict'):
    """
    Similar to six.ensure_str. Adapted here to avoid messing up with six version errors.
    """
    if not PYTHON3 and isinstance(s, text_type):
        return s.encode(encoding, errors)
    elif PYTHON3 and isinstance(s, binary_type):
        try:
            return s.decode(encoding, errors)
        except:
            return s.decode("latin-1")
    elif not isinstance(s, (text_type, binary_type)):
        raise TypeError("not expecting type '%s'" % type(s))
    return s


def execfile(source, globals=None, locals=None):
    with open(source) as f:
        content = f.read()
    exec(content, globals, locals)
if PYTHON3:
    __all__ += ["execfile"]


def iterbytes(text):
    """
    Bytes iterator. If a string is provided, it will automatically be converted to bytes.
    """
    if isinstance(text, string_types):
        text = b(text)
    for c in text:
        yield c if PYTHON3 else ord(c)




# -*- coding: UTF-8 -*-
"""Network-related checking functions and argument types.

"""
import netaddr
import netifaces
import re
from email.utils import parseaddr as parse_email
from itertools import chain

from .strings import _str2list


__all__ = __features__ = []


# network-related check functions
__all__ += ["is_asn", "is_defgw", "is_domain", "is_email", "is_gw", "is_hostname", "is_ifaddr", "is_ip", "is_ipv4",
            "is_ipv6", "is_mac", "is_netif", "is_port", "is_url"]
is_asn      = lambda a:   __as_number(a, False) is not None
is_defgw    = lambda gw:  __gateway_address(gw, True, False) is not None
is_domain   = lambda n:   __domain_name(n, False, False) is not None
is_email    = lambda e:   __email_address(e, False) is not None
is_gw       = lambda gw:  __gateway_address(gw, False, False) is not None
is_hostname = lambda h:   __hostname(h, False) is not None
is_ifaddr   = lambda ip:  __interface_address(ip, False) is not None
is_ip       = lambda ip:  __ip_address(ip, None, False) is not None
is_ipv4     = lambda ip:  __ip_address(ip, 4, False) is not None
is_ipv6     = lambda ip:  __ip_address(ip, 6, False) is not None
is_mac      = lambda mac: __mac_address(mac, False) is not None
is_netif    = lambda nif: __network_interface(nif, False) is not None
is_port     = lambda p:   isinstance(p, int) and 0 < p < 2**16
is_url      = lambda url: __url(url, False) is not None


# network-related argument types
__all__ += ["as_number", "default_gateway_address", "domain_name", "email_address", "gateway_address", "hostname",
            "ip_address", "ipv4_address", "ipv6_address", "ip_address_list", "ipv4_address_list", "ipv6_address_list",
            "ip_address_filtered_list", "ipv4_address_filtered_list", "ipv6_address_filtered_list",
            "ip_address_network", "ipv4_address_network", "ipv6_address_network", "interface_address",
            "interface_address_list", "interface_address_filtered_list", "mac_address", "network_interface",
            "port_number", "port_number_range", "url"]


def __as_number(asn, fail=True):
    """ Autonomous System Number validation. """
    if str(asn).isdigit() and 0 <= int(asn) < 2**32:
        return asn
    if fail:
        raise ValueError("Bad AS number")
as_number = lambda a: __as_number(a)


def __domain_name(name, dotted=False, fail=True):
    """ Domain name validation. """
    if not dotted and not name.endswith("."):
        name += "."
    # source: https://stackoverflow.com/questions/19705002/
    if len(name) <= 255 and re.match(r"^(((([a-z0-9]+){1,63}\.)|(([a-z0-9]+"
                r"(\-)+[a-z0-9]+){1,63}\.))+){1,255}$", name, re.I) is not None:
        return name
    if fail:
        raise ValueError("Bad domain name")
domain_name = lambda n: __domain_name(n)


def __email_address(email, fail=True):
    """ Email address validation. """
    # reference: https://stackoverflow.com/questions/8022530/
    if len(email) <= 320 and re.match(r"^[^@]+@[^@]+$", email) and \
       is_hostname(email.split("@")[1]) and parse_email(email)[1] != "":
        return email
    if fail:
        raise ValueError("Bad email address")
email_address = lambda e: __email_address(e)


def __gateway_address(gw, default=False, fail=True):
    """ Gateway IP address validation. """
    g = netifaces.gateways() if not default else netifaces.gateways()['default']
    for k, v in g.items():
        if k == "default":
            continue
        if isinstance(v, tuple):
            v = [v]
        for l in v:
            if gw == l[0]:
                return gw
    if fail:
        raise ValueError("Bad {}gateway".format(["", "default "][default]))
default_gateway_address = lambda gw: __gateway_address(gw, True)
gateway_address         = lambda gw: __gateway_address(gw)


HN = re.compile(r"(?!-)[A-Z\d-]{1,63}(?<!-)$", re.I)
def __hostname(hostname, fail=True):
    """ Hostname validation. """
    if len(hostname) <= 255 and all(HN.match(x) for x in \
       hostname.rstrip(".").split(".")):
        return hostname
    if fail:
        raise ValueError("Bad hostname")
hostname = lambda h: __hostname(h)


def __interface_address(addr, fail=True):
    """ Interface address validation. """
    for iface in netifaces.interfaces():
        for k, v in netifaces.ifaddresses(iface).items():
            for d in v:
                if addr in d.values():
                    return addr
    if fail:
        raise ValueError("Bad interface address")
interface_address = lambda a: __interface_address(a)


def __interface_address_list(addrs, filter_bad=False):
    """ Interface addresses validation and expansion. """
    addrs = _str2list(addrs)
    l = []
    for addr in addrs:
        try:
            l.append(__interface_address(addr))
        except ValueError as e:
            if not filter_bad:
                raise ValueError(str(e))
    return l
interface_address_list          = lambda a: __interface_address_list(a)
interface_address_filtered_list = lambda a: __interface_address_list(a, True)


def __ip_address(ip, version=None, fail=True):
    """ IP address validation. """
    # note: netaddr already handles validation and raises a ValueError in case
    #        of bad address ; we just ensure that the input is converted to
    #        unicode using six.u (otherwise, it fails in Python 2)
    ip = int(ip) if str(ip).isdigit() else str(ip)
    try:
        return netaddr.IPAddress(ip, version=version)
    except (ValueError, netaddr.core.AddrFormatError) as e:
        if fail:
            raise ValueError(str(e))
ip_address   = lambda ip: __ip_address(ip)
ipv4_address = lambda ip: __ip_address(ip, 4)
ipv6_address = lambda ip: __ip_address(ip, 6)


def __ip_address_list(ips, version=None, filter_bad=False):
    """ IP address range validation and expansion. """
    ips = _str2list(ips)
    # consider it as an ipaddress.IPv[4|6]Network instance and expand it
    l = []
    for ip in ips:
        # parse it as a single IP
        _ = __ip_address(ip, version, False)
        if _ is not None and _ not in l:
            l.append([_])
            continue
        # parse it as a network
        try:
            l.append(netaddr.IPNetwork(ip, version=version))
        except (ValueError, netaddr.core.AddrFormatError) as e:
            if not filter_bad:
                raise ValueError(str(e))
    # make a generator with the parsed IP addresses/networks
    def __generator():
        _ = []
        for ip in chain(*l):
            if ip not in _:
                yield str(ip)
                _.append(ip)
    if len(l) > 0:
        return __generator()
ip_address_list            = lambda ips: __ip_address_list(ips)
ip_address_filtered_list   = lambda ips: __ip_address_list(ips, None, True)
ipv4_address_list          = lambda ips: __ip_address_list(ips, 4)
ipv4_address_filtered_list = lambda ips: __ip_address_list(ips, 4, True)
ipv6_address_list          = lambda ips: __ip_address_list(ips, 6)()
ipv6_address_filtered_list = lambda ips: __ip_address_list(ips, 6, True)
ip_address_network         = lambda net: __ip_address_list([net], None)
ipv4_address_network       = lambda net: __ip_address_list([net], 4)
ipv6_address_network       = lambda net: __ip_address_list([net], 6)


def __mac_address(mac, fail=True):
    """ MAC address validation. """
    msg = "'{}' does not appear to be a MAC address".format(mac)
    # check as an integer
    if str(mac).isdigit():
        mac = int(mac)
    try:
        return netaddr.EUI(mac)
    except (ValueError, netaddr.core.AddrFormatError) as e:
        if fail:
            raise ValueError(str(e))
        else:
            return
mac_address = lambda mac: __mac_address(mac)


def __network_interface(netif, fail=True):
    """ Network interface validation. """
    if netif in netifaces.interfaces():
        return netif
    if fail:
        raise ValueError("Not an existing interface")
network_interface = lambda nif: __network_interface(nif)


def __url(url, fail=True):
    """ URL hyperlink. """
    if len(url) > 2048:
        raise ValueError("Not a valid URL (too long)")
    try:
        scheme, link = url.split("://")
    except ValueError:
        if fail:
            raise ValueError("Not a valid URL (cannot parse scheme)")
        else:
            return
    if len(scheme) > 36:
        if fail:
            raise ValueError("Not a valid URL (invalid scheme)")
        else:
            return
    link = link.split("?")
    link, query = link if len(link) == 2 else (link[0], "")
    link = link.split("/")
    link, reqpath = link if len(link) == 2 else (link[0], "")
    link = link.split("@")
    auth, domain = link if len(link) == 2 else ("", link[0])
    if not is_domain(domain):
        if fail:
            raise ValueError("Not a valid URL (bad domain)")
        else:
            return
    auth = auth.split(":")
    if auth != [''] and (len(auth) != 2 or auth[0] == "" and auth[1] == ""):
        if fail:
            raise ValueError("Not a valid URL (bad auth)")
        else:
            return
    return url
url = lambda url: __url(url)


def port_number(port):
    """ Port number validation. """
    try:
        port = int(port)
    except ValueError:
        raise ValueError("Bad port number")
    if not 0 <= port < 2 ** 16:
        raise ValueError("Bad port number")
    return port


def port_number_range(prange):
    """ Port number range validation and expansion. """
    # first, try it as a normal port number
    try:
        return port_number(prange)
    except ValueError:
        pass
    # then, consider it as a range with the format "x-y" and expand it
    try:
        bounds = list(map(int, re.match(r'^(\d+)\-(\d+)$', prange).groups()))
        if bounds[0] > bounds[1]:
            raise AttributeError()
    except (AttributeError, TypeError):
        raise ValueError("Bad port number range")
    return list(range(bounds[0], bounds[1] + 1))



# -*- coding: UTF-8 -*-
"""Module for text-related utility functions.

"""
import mdv
import re
from gettext import gettext as gt
from pypandoc import convert_text

from .data.types.network import is_email, is_url


__features__ = ["gt", "txt2blockquote", "txt2bold", "txt2email", "txt2italic", "txt2olist", "txt2paragraph",
                "txt2title", "txt2ulist", "txt2underline", "txt2url"]
__all__ = __features__ + ["DOCFORMAT_THEME"]

DOCFORMAT = None
DOCFORMAT_THEME = "Makeup"
FORMATS = [None, "html", "md", "rst", "textile"]


def __check(**kwargs):
    for k, v in kwargs.items():
        if k == 'email' and not is_email(v):
            raise ValueError("Invalid email address")
        elif k == 'format' and v not in FORMATS:
            raise ValueError("Invalid format ; should be one of ({})".format("|".join(map(str, FORMATS))))
        elif k == 'level' and v not in range(1, 7):
            raise ValueError("Invalid title level ; should belong to [1,6]")
        elif k in ['bold', 'italic', 'ordered', 'underline'] and not isinstance(v, bool):
            raise ValueError("Bad boolean value")
        elif k == 'url' and not is_url(v):
            raise ValueError("Invalid URL")


def configure_docformat(glob):
    """ This reconfigures the globally selected documentation format.
    
    :param glob: globals dictionary
    """
    format = glob.get('__docformat__')
    __check(format=format)
    global DOCFORMAT, DOCFORMAT_THEME
    DOCFORMAT = format
    DOCFORMAT_THEME = glob.get('DOCFORMAT_THEME', DOCFORMAT_THEME)


def txt_terminal_render(text, format=None):
    """ This renders input text based on the selected format.
    
    :param format: selected format (one of FORMATS)
    """
    format = format or DOCFORMAT
    __check(format=format)
    if format is None:
        return text
    # collect whitespaces in argument lines and line indentations
    ARG_REGEX = re.compile(r"((?P<h>\-{1,2})[a-z][a-z0-9_]*(?:\s+[A-Z][A-Z_]*)?"
                           r"(?:\,\s+(?P=h)\-[a-z][a-z0-9_]*(?:\s+[A-Z][A-Z_]*)?)?)(\s+)(.*)$")
    spaces = {}
    for line in text.split("\n"):
        s = ARG_REGEX.search(line)
        if s:
            spaces[s.group(1)] = len(s.group(3))
            # e.g. "-v, --verbose" and 10 (whitespaces)
        else:
            tmp = line.lstrip()
            indent = len(line) - len(tmp)
            if indent > 0:
                spaces[tmp.strip()] = indent
    # convert here   
    md = text if format == "md" else convert_text(text, "md", format=format).replace("\\", "")
    if format != "md":
        # bug corrections
        # 1. misaligned metadata fields
        # 2. rectify misaligned arguments' helps
        tmp = md
        md, l = "", 0
        #  collect the maximum field length
        for line in tmp.split("\n"):
            try:
                f, v = line.split(":", 1)
                if f.endswith("default"):
                    raise ValueError
                if re.search(r"(?P<t>[_\*])(.*?)(?P=t)", f):
                    f = "_" + f.strip(" *_") + "_"
                    l = max(l, len(f))
            except ValueError:
                pass
        #  then, remake markdown string
        for line in tmp.split("\n"):
            try:
                f, v = line.split(":", 1)
                if f.endswith("default"):
                    raise ValueError()
                if re.search(r"(?P<t>[_\*])(.*?)(?P=t)", f):
                    f = "_" + f.strip(" *_") + "_"
                    md += ("{: <%d}: {}" % l).format(f, v) + "\n"
                    continue
            except ValueError:
                pass
            # restore spaces between arguments and their help messages
            s = ARG_REGEX.search(line)
            if s:
                l = spaces[s.group(1)]
                line = ARG_REGEX.sub(r"\1{}\4".format(" " * l), line)
            # restore indentations
            else:
                try:
                    line = spaces[line] * " " + line
                except KeyError:
                    pass
            md += line + "\n"
        # 3. links incorrectly rendered with mdv after using pandoc
        for link in re.findall("(<(.*?)>)", md):
            md = md.replace(link[0], "[{0}]({1}{0})".format(link[1], ["", "mailto:"][is_email(link[1])]))
    return mdv.main(md, display_links=True, theme=DOCFORMAT_THEME)


def _txt_list(text, format=None, ordered=False):
    """ This reformats an input raw text list based on the selected format.
    
    :param format:  selected format (one of FORMATS)
    :param ordered: whether the list is ordered or not
    """
    format = format or DOCFORMAT
    __check(format=format, ordered=ordered)
    r = text
    if format == "html":
        r = "<{}l>".format(["u", "o"][ordered])
    elif format in ["md", "rst", "textile"]:
        r = ""
    for line in re.split("\n[\-\*] ", text):
        line = line.lstrip("*- ")
        if format == "html":
            r += "\n<li>{}</li>".format(line)
        elif format in ["md", "rst"]:
            r += "\n- {}".format(line)
        elif format == "textile":
            r += "\n{} {}".format(["*", "#"][ordered], line)
    if format == "html":
        r += "\n</{}l>".format(["u", "o"][ordered])
    return r.strip("\n")
txt2olist = lambda text, format=None: _txt_list(text, format, True)
txt2ulist = lambda text, format=None: _txt_list(text, format)


def _txt_style(text, format=None, bold=False, italic=False, underline=False):
    """ This restyles an input raw text based on the selected format.
    
    :param format:    selected format (one of FORMATS)
    :param bold:      whether the text should be bold
    :param italic:    whether the text should be italic
    :param underline: whether the text should be underline
    """
    format = format or DOCFORMAT
    __check(format=format, bold=bold, italic=italic, underline=underline)
    if format == "html":
        if bold:
            text = "<b>{}</b>".format(text)
        if italic:
            text = "<i>{}</i>".format(text)
        if underline:
            text = "<u>{}</u>".format(text)
    elif format == "md":
        if bold:
            text = "**{}**".format(text)
        if italic:
            text = "_{}_".format(text)
        if underline:  # note: this is a ugly trick using hyperlink
            text = "[{}]()".format(text)
    elif format in ["rst", "textile"]:
        if bold:
            text = "**{}**".format(text)
        if italic:
            text = "*{}*".format(text)
        if underline:
            text = "_{}_".format(text)
    return text
txt2bold = lambda text, format=None: _txt_style(text, format, bold=True)
txt2italic = lambda text, format=None: _txt_style(text, format, italic=True)
txt2underline = lambda text, format=None: _txt_style(text, format, underline=True)


def txt2blockquote(text, format=None):
    """ This reformats a text to a block quote based on the selected format.
    
    :param format: selected format (one of FORMATS)
    """
    format = format or DOCFORMAT
    __check(format=format)
    text = text.strip("\n")
    if format == "html":
        text = "<blockquote>{}</blockquote>".format(text.strip("\n"))
    elif format == "md":
        text = "\n".join("> {}".format(line) for line in text.splitlines())
    elif format == "rst":
        text = "\n".join("\t{}".format(line) for line in text.splitlines())
    elif format == "textile":
        text = "\n".join("" if line.strip() == "" else "bq. " + line for line in text.splitlines())
    return text


def txt2email(text, format=None):
    """ This reformats an email to a hyperlink based on the selected format.
    
    :param format: selected format (one of FORMATS)
    :param text:   email address
    """
    format = format or DOCFORMAT
    __check(format=format, email=text)
    if format == "html":
        text = "<a href=\"mailto:{0}\">{0}</a>".format(text)
    elif format == "md":
        text = "[{0}](mailto:{0})".format(text)
    elif format == "textile":
        text = "\"{0}\":mailto:{0}".format(text)
    return text


def txt2paragraph(text, format=None):
    """ This reformats a text to a paragraph based on the selected format.
    
    :param format: selected format (one of FORMATS)
    """
    format = format or DOCFORMAT
    __check(format=format)
    if format == "html":
        text = "<p>{}</p>".format(text)
    elif format in ["rst", "textile"]:
        text = "\n" + text
    return text


def txt2title(text, format=None, level=2):
    """ This reformats a text to a title based on the selected format.
    
    :param format: selected format (one of FORMATS)
    :param level:  title level
    """
    format = format or DOCFORMAT
    __check(format=format, level=level)
    if format is not None:
        text = text.title()
    if format == "html":
        text = "<h{0}>{1}</h{0}>".format(level, text)
    elif format == "md":
        text = "{} ".format(level * "#") + text
    elif format == "rst":
        text = "{}\n{}".format(text, len(text) * "=-`'~*"[level - 1])
    elif format == "textile":
        text = "h{}. {}".format(level, text)
    return text


def txt2url(text, format=None, url=None):
    """ This reformats a text to a hyperlink based on the selected format.
    
    :param format: selected format (one of FORMATS)
    :param url:    URL
    """
    format = format or DOCFORMAT
    if url is None:
        __check(format=format, url=text)
        url = text
    else:
        __check(format=format, url=url)
    if format == "html":
        text = "<a href=\"{}\">{}</a>".format(url, text)
    elif format == "md":
        text = "[{}]({})".format(text, url)
    elif format == "textile":
        if text == url:
            text = "$"
        text = "\"{}\":{}".format(text, url)
    return text




# -*- coding: UTF-8 -*-
"""Files/Folders-related checking functions and argument types.

"""
from os import access, makedirs, X_OK
from os.path import exists, isdir, isfile

from .strings import _str2list


__all__ = __features__ = []


# dummy shortcuts, compliant with the is_* naming convention
__all__ += ["is_dir", "is_executable", "is_file", "is_folder"]
is_dir = is_folder = isdir
is_executable = lambda f: access(f, X_OK)
is_file = isfile


# file and folder-related argument types
__all__ += ["file_does_not_exist", "file_exists", "files_list",
            "files_filtered_list", "folder_does_not_exist", "folder_exists",
            "folder_exists_or_create"]


def file_does_not_exist(f):
    """ Check that the given file does not exist. """
    if exists(f):
        raise ValueError("'{}' already exists".format(f))
    return f
folder_does_not_exist = file_does_not_exist


def file_exists(f):
    """ Check that the given file exists. """
    if not exists(f):
        raise ValueError("'{}' does not exist".format(f))
    if not isfile(f):
        raise ValueError("Target exists and is not a file")
    return f


def files_list(l, filter_bad=False):
    """ Check if the list contains valid files. """
    l = _str2list(l)
    nl = []
    for f in l:
        if not isfile(f):
            if not filter_bad:
                raise ValueError("A file from the given list does not exist")
        else:
            nl.append(f)
    if filter_bad and len(nl) == 0:
        raise ValueError("No valid file in the given list")
    return nl


def files_filtered_list(l):
    """ Check if the list contains valid files and discard invalid ones. """
    return files_list(l, True)


def folder_exists(f):
    """ Check that the given folder exists. """
    if not exists(f):
        raise ValueError("'{}' does not exist".format(f))
    if not isdir(f):
        raise ValueError("Target exists and is not a folder")
    return f


def folder_exists_or_create(f):
    """ Check that the given folder exists and create it if not existing. """
    if not exists(f):
        makedirs(f)
    if not isdir(f):
        raise ValueError("Target exists and is not a folder")
    return f



# -*- coding: UTF-8 -*-
"""String-related checking functions and argument types.

"""
import ast
import re
from six import binary_type, string_types
from string import ascii_lowercase as LC, ascii_uppercase as UC, ascii_letters as letter, digits, printable, punctuation


__all__ = __features__ = []


def _str2list(s):
    """ Convert string to list if input is effectively a string. """
    # if already a list, simply return it, otherwise ensure input is a string
    if isinstance(s, list):
        return s
    else:
        s = str(s)
    # remove heading and trailing brackets
    if s[0] == '[' and s[-1] == ']' or s[0] == '(' and s[-1] == ')':
        s = s[1:-1]
    # then parse list elements from the string
    l = []
    for i in s.split(","):
        i = i.strip()
        try:
            l.append(ast.literal_eval(i))
        except Exception:
            l.append(i)
    return l


def _is_from_alph(s, a, t):
    if is_str(s):
        val = str_contains(a, t)
        try:
            val(s)
            return True
        except ValueError:
            pass
    return False


# various string-related check functions
__all__ += ["is_str", "is_bytes", "is_digits", "is_letters", "is_lowercase", "is_printable", "is_punctuation",
            "is_uppercase"]
is_str         = lambda s: isinstance(s, string_types)
is_bytes       = lambda s: isinstance(s, binary_type)
is_digits      = lambda s, t=1.0: _is_from_alph(s, digits, t)
is_letters     = lambda s, t=1.0: _is_from_alph(s, letter, t)
is_lowercase   = lambda s, t=1.0: _is_from_alph(s, LC, t)
is_printable   = lambda s, t=1.0: _is_from_alph(s, printable, t)
is_punctuation = lambda s, t=1.0: _is_from_alph(s, punctuation, t)
is_uppercase   = lambda s, t=1.0: _is_from_alph(s, UC, t)

# various data format check functions
__all__ += ["is_bin", "is_hex"]
is_bin = lambda b: is_str(b) and all(set(_).difference(set("01")) == set() for _ in re.split(r"\W+", b)) or \
                   isinstance(b, (list, set, tuple)) and all(str(x) in "01" for x in b)
is_hex = lambda h: is_str(h) and len(h) % 2 == 0 and set(h.lower()).difference(set("0123456789abcdef")) == set()

# some other common check functions
__all__ += ["is_long_opt", "is_short_opt"]
is_long_opt  = lambda o: is_str(o) and re.match(r"^--[a-z]+(-[a-z]+)*$", o, re.I)
is_short_opt = lambda o: is_str(o) and re.match(r"^-[a-z]$", o, re.I)


# -------------------- STRING FORMAT ARGUMENT TYPES --------------------
__all__ += ["str_contains", "str_matches"]


def str_contains(alphabet, threshold=1.0):
    """ Counts the characters of a string and determines, given an alphabet, if the string has enough valid characters.
    """
    if threshold < 0.0 or threshold > 1.0:
        raise ValueError("Bad threshold (should be between 0 and 1)")
    def _validation(s):
        p = sum(int(c in alphabet) for c in s) / float(len(s))
        if p < threshold:
            raise ValueError("Input string does not contain enough items from the given alphabet ({:.2f}%)"
                             .format(p * 100))
        return s
    return _validation


def str_matches(pattern, flags=0):
    """ Applies the given regular expression to a string argument. """
    def _validation(s):
        if re.match(pattern, s, flags) is None:
            raise ValueError("Input string does not match the given regex")
        return s
    return _validation




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Module for defining logging-related constants and objects.

"""
import coloredlogs
from datetime import timedelta
from time import gmtime

from .preimports import logging


__features__ = ["LOG_FORMAT", "DATE_FORMAT", "TIME_MILLISECONDS", "logger"]
__all__      = ["coloredlogs", "configure_logger"] + __features__


DATE_FORMAT       = '%H:%M:%S'
LOG_FORMAT        = '%(asctime)s [%(levelname)s] %(message)s'
TIME_MILLISECONDS = False


# add a custom log level for interactive mode
logging.addLogLevel("interact", "cyan", 100)
# add a custom log level for stepping
logging.addLogLevel("step", "cyan", 101)
# add a custom log level for timing
logging.addLogLevel("time", "magenta", 102)
# add a custom success log level
logging.addLogLevel("success", "green", 103)
# add a custom failure log level
logging.addLogLevel("failure", "red", 104)


# setup a default logger for allowing logging before initialize() is called
logger = logging.getLogger("main")
logger.setLevel(1)
logger.addHandler(logging.InterceptionHandler())
coloredlogs.DEFAULT_LOG_FORMAT = LOG_FORMAT
coloredlogs.DEFAULT_DATE_FORMAT = DATE_FORMAT
coloredlogs.install(logger=logger)


class RelativeTimeColoredFormatter(coloredlogs.ColoredFormatter):
    """
    Custom formatter for computing relative times.
    """
    converter = gmtime
    
    def __init__(self, *args, **kwargs):
        super(RelativeTimeColoredFormatter, self).__init__(*args, **kwargs)
        self.datefmt = '%H:%M:%S.%f'
    
    def format(self, record):
        record.created = timedelta(microseconds=record.relativeCreated).total_seconds()
        return super(RelativeTimeColoredFormatter, self).format(record)


def configure_logger(glob, multi_level, relative=False, logfile=None, syslog=False):
    """
    Logger configuration function for setting either a simple debug mode or a multi-level one.
    
    :param glob:        globals dictionary
    :param multi_level: boolean telling if multi-level debug is to be considered
    :param relative:    use relative time for the logging messages
    :param logfile:     log file to be saved (None means do not log to file)
    :param syslog:      enable logging to /var/log/syslog
    """
    _l = logging
    levels = [_l.ERROR, _l.WARNING, _l.INFO, _l.DEBUG] if multi_level else [_l.INFO, _l.DEBUG]
    try:
        verbose = min(int(glob['args'].verbose), 3)
    except AttributeError:
        verbose = 0
    glob['args']._debug_level = dl = levels[verbose]
    glob['args']._debug_syslog = syslog
    glob['args']._debug_logfile = logfile
    glob['logger'] = logger
    logger.handlers = []
    logger.setLevel(1)
    logger.addHandler(_l.InterceptionHandler())
    handler = _l.StreamHandler()
    lfmt = "\r" + glob.get('LOG_FORMAT', LOG_FORMAT)
    dfmt = glob.get('DATE_FORMAT', DATE_FORMAT)
    formatter = _l.Formatter(lfmt, dfmt)
    handler.setFormatter(formatter)
    handler.setLevel(dl)
    logger.addHandler(handler)
    if relative:
        coloredlogs.ColoredFormatter = RelativeTimeColoredFormatter
    gtms = glob.get('TIME_MILLISECONDS', TIME_MILLISECONDS)
    coloredlogs.install(dl, logger=logger, fmt=lfmt, datefmt=dfmt, syslog=syslog, stream=logfile, milliseconds=gtms)
    lastrec = _l.getLogger("__last_record__")
    lastrec.handlers = []
    handler = _l.StreamHandler()
    handler.setFormatter(_l.Formatter("\r" + lfmt, dfmt))
    lastrec.addHandler(handler)
    lastrec.setLevel(1)
    coloredlogs.install(1, logger=lastrec, fmt="\r" + lfmt, datefmt=dfmt, milliseconds=gtms)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Module for defining handlers and exit hook.

"""
import sys
from signal import getsignal, siginterrupt, signal, \
                   SIG_IGN, SIGINT, SIGTERM, SIGUSR1

from .helpers.inputs import user_input


__features__ = ["at_exit", "at_graceful_exit", "at_interrupt", "at_terminate",
                "DisableSignals"]
__all__ = ["_hooks"] + __features__


class DisableSignals(object):
    """
    Context manager that disable signal handlers.

    :param signals: list of signal identifiers
    :param fail:    whether execution should fail or not when a bad signal ID is
                     encountered
    """
    def __init__(self, *signals, **kwargs):
        self.__handlers = {}
        for s in signals:
            try:
                self.__handlers[s] = getsignal(s)
            except ValueError as e:
                if kwargs.get('fail', False):
                    raise e

    def __enter__(self):
        for s in self.__handlers.keys():
            signal(s, SIG_IGN)

    def __exit__(self, exc_type, exc_val, exc_tb):
        for s, h in self.__handlers.items():
            signal(s, h)


class ExitHooks(object):
    sigint_actions = ["confirm", "continue", "exit"]
    # inspired from: https://stackoverflow.com/questions/9741351/how-to-find-exi
    #                 t-code-or-reason-when-atexit-callback-is-called-in-python
    def __init__(self):
        self.__sigint_action = "exit"
        self._orig_exit = sys.exit
        self.code = None
        self.exception = None
        sys.exit = self.exit
        self.resume()
    
    def exit(self, code=0):
        self.code = code
        self._orig_exit(code)
    
    def pause(self):
        self.state = "PAUSED"
        while self.state == "PAUSED": continue
    
    def quit(self, code=0):
        if self.__sigint_action == "confirm" and \
           user_input("Do you really want to interrupt execution ?",
                      ["(Y)es", "(N)o"], "y", style="bold") == "yes":
            self.__sigint_action = "exit"
        if self.state != "INTERRUPTED" or self.__sigint_action == "exit":
            self.exit(code)
        self.resume()
    
    def resume(self):
        self.state = "RUNNING"
    
    @property
    def sigint_action(self):
        return self.__sigint_action
    
    @sigint_action.setter
    def sigint_action(self, value):
        if value not in self.sigint_actions:
            raise ValueError("Bad interrupt action ; should be one of {}"
                             .format("|".join(self.sigint_actions)))
        self.__sigint_action = value

_hooks = ExitHooks()


def __interrupt_handler(*args):
    """
    Interruption handler.

    :param signal: signal number
    :param stack: stack frame
    :param code: exit code
    """
    _hooks.state = "INTERRUPTED"
    _hooks.quit(0)
# bind to interrupt signal (Ctrl+C)
signal(SIGINT, __interrupt_handler)


def __pause_handler(*args):
    """
    Execution pause handler.
    """
    _hooks.pause()
# bind to user-defined signal
signal(SIGUSR1, __pause_handler)
siginterrupt(SIGUSR1, False)


def __terminate_handler(*args):
    """
    Termination handler.

    :param signal: signal number
    :param stack: stack frame
    :param code: exit code
    """
    _hooks.state = "TERMINATED"
    _hooks.quit(0)
# bind to terminate signal
signal(SIGTERM, __terminate_handler)


at_exit          = lambda: None
at_graceful_exit = lambda: None
at_interrupt     = lambda: None
at_terminate     = lambda: None



# -*- coding: UTF-8 -*-
"""Module for defining Report.

"""
import logging
from os import listdir
from os.path import abspath, dirname, isfile, join, splitext
try:
    from weasyprint import CSS, HTML
    pdf_generation = True
except ImportError:
    pdf_generation = False

from .base import *
from .objects import *
from ..helpers import ensure_str
from ..preimports import logging


__all__ = __features__ = ["Report"]

PAGE_CSS = "@page{size:%(size)s;margin:%(margins)s;%(header)s%(footer)s}"


class Report(list):
    """ This class represents a whole report. """
    page_css = "@page{size:%(size)s;margin:%(margins)s;%(header)s%(footer)s}"
    
    @logging.bindLogger
    def __init__(self, *pieces, **options):
        self.filename = options.get('filename', "report") or "report"
        self.size = options.get('size', "a4 portrait")
        self.margins = options.get('margins', "10.0mm 20.0mm 20.0mm 20.0mm")
        self.theme = options.get('theme', "default")
        if self.theme not in self.themes:
            raise ValueError("PDF report CSS theme does not exist")
        self.css = options.get('css', join(dirname(abspath(__file__)), "{}.css".format(self.theme)))
        self.css = abspath(self.css)
        if not isfile(self.css):
            raise ValueError("PDF report CSS file does not exist")
        self.title = options.get('title')
        self.clear()
        self.append(*pieces)
    
    def __handle(self, piece):
        p = piece
        if p.__class__.__name__ == "Footer":
            if self.footer:
                return
            self.footer = p.css
        elif p.__class__.__name__ == "Header":
            if self.header:
                return
            self.header = p.css
        elif not isinstance(p, Element):
            p = Text(str(p))
        c = p.__class__.__name__.lower()
        self.__counts.setdefault(c, 0)
        self.__counts[c] += 1
        p.id = self.__counts[c]
        if self.__counts[c] > 1:
            if self.__counts[c] == 2:
                for p2 in self:
                    if p2.__class__.__name__.lower() == p2.name == c:
                        p2.name += "-%d" % p2.id
            if p.name == c:
                p.name += "-%d" % p.id
        return p
    
    def append(self, *pieces):
        for p in pieces:
            p = self.__handle(p)
            if p:
                super(Report, self).append(p)
    
    def clear(self):
        self.__counts = {}
        self.header = ""
        self.footer = ""
        if self.title:
            self.insert(0, Title(self.title))
    
    def copy(self):
        r = Report(**self.__dict__)
        r.append(*self)
        return r
    
    def count(self, piece_class):
        return self.__counts[piece_class.lower()]
    
    def extend(self, pieces):
        self.append(*pieces)
    
    def index(self, name):
        for p in self:
            if p.name == name:
                return p
        raise ValueError("'{}' is not in list".format(name))
    
    def insert(self, index, piece):
        p = self.__handle(piece)
        if p:
            super(Report, self).insert(index, p)
    
    def prepend(self, *pieces):
        for p in pieces[::-1]:
            p = self.__handle(p)
            if p:
                super(Report, self).insert(0, p)
    
    @property
    def counts(self):
        return self.__counts
    
    @property
    def themes(self):
        return list(map(lambda f: splitext(f)[0], filter(lambda f: f.endswith(".css"), listdir(dirname(__file__)))))
    
    @output
    def csv(self, text=TEXT):
        return "\n\n".join(p.csv(True) for p in self if p.csv(True) != "")
    
    @output
    def html(self, indent=4, text=TEXT):
        """ Generate an HTML file from the report data. """
        ind = (indent or 0) * " "
        self.logger.debug("Generating the HTML report{}...".format(["", " (text only)"][text]))
        r = []
        for p in self:
            h = p.html(indent=indent, text=True)
            if h != "":
                r.append(h)
        nl = "" if indent is None else "\n"
        h = (nl + "<br>" + nl).join(r).replace("\\\"", "\"")
        r = []
        for line in h.split("\n"):
            r.append(2 * ind + line)
        r = ["<html>", ind + "<body>"] + r + [ind + "</body>", "</html>"]
        return nl.join(r)
    
    @output
    def json(self, data_only=False, text=TEXT):
        r = {}
        for p in self:
            if isinstance(p, (Data, List, Table)) or not data_only:
                r.update(p.json(True))
        return r
    
    @output
    def md(self, text=TEXT):
        r = []
        for p in self:
            if p.md() != "":
                r.append(p.md())
        return "\n\n".join(r)

    @output
    def pdf(self, text=TEXT):
        """ Generate a PDF file from the report data. """
        if pdf_generation:
            self.logger.debug("Generating the PDF report...")
            html = HTML(string=self.html())
            fn = self.filename if self.filename.endswith(".pdf") else "%s.pdf" % self.filename
            html.write_pdf(fn, stylesheets=[self.css, CSS(string=PAGE_CSS % self.__dict__)])
    
    @output
    def xml(self, indent=2, data_only=False, text=TEXT):
        r = ["<report>"]
        for p in self:
            if isinstance(p, (Data, List, Table)) or not data_only:
                out = ensure_str(p.xml(indent=indent, text=True))
                for line in out.split("\n"):
                    r.append((indent or 0) * " " + line)
        r.append("</report>")
        r = ("" if indent is None else "\n").join(r)
        return r




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Module for defining the notify feature.

"""
from os.path import dirname, isfile, join

from ..helpers.common import notify
from ..helpers.constants import LINUX
from ..helpers.data.types import folder_exists, positive_int
from ..preimports import logging


__all__ = ["set_notify_items"]


def set_notify_items(glob):
    """
    This function prepares the notify items for inclusion in main script's global scope.
    
    :param glob: main script's global scope dictionary reference
    """
    a = glob['args']
    enabled = getattr(a, a._collisions.get("notify") or "notify", False)
    appname = glob.get('__scriptname__')
    timeout = positive_int(glob.get('NOTIFICATION_TIMEOUT', 5), zero=False)
    icon_path = folder_exists(glob.get('NOTIFICATION_ICONS_PATH', dirname(__file__)))
    level = positive_int(glob.get('NOTIFICATION_LEVEL', logging.SUCCESS))
    
    class __NotificationHandler(logging.Handler):
        def emit(self, record):
            title = "{}[{}]:".format(appname, record.name) if record.name != "main" else appname
            icon = record.levelname.lower()
            ipath = join(icon_path, icon) + ".png"
            if isfile(ipath):
                icon = ipath
            print(title, record.msg, appname, icon, timeout, record.levelname)
            notify(title, record.msg, appname, icon, timeout, title + " " + record.levelname)
    
    if enabled and not any(type(h) is __NotificationHandler for h in glob['logger'].handlers):
        nh = __NotificationHandler()
        nh.setLevel(level)
        glob['logger'].addHandler(nh)
    glob['notify'] = notify



# -*- coding: UTF-8 -*-
"""Module containing various helper functions.

"""
from pprint import pprint
from types import ModuleType

from .classprop import *
from .common import *
from .compat import *
from .constants import *
from .data import *
from .decorators import *
from .fexec import *
from .inputs import *
from .licenses import *
from .path import *
from .termsize import *
from .text import *
from .timeout import *

from .classprop import __features__ as _classprop
from .common import __features__ as _common
from .compat import __features__ as _compat
from .constants import __features__ as _constants
from .data import __features__ as _data
from .decorators import __features__ as _decorators
from .fexec import __features__ as _fexec
from .inputs import __features__ as _inputs
from .licenses import __features__ as _licenses
from .path import __features__ as _path
from .termsize import __features__ as _termsize
from .text import __features__ as _text
from .timeout import __features__ as _timeout


__helpers__ = _common + _data + _decorators + _fexec + _inputs + _licenses + _path + _termsize + _text + _timeout

ts = ModuleType("ts", """
    Tinyscript helpers
    ~~~~~~~~~~~~~~~~~~

    The `ts` module contains various helper functions that can be very useful
     to not reinvent the wheel, including:
    
    - Common utility functions
    
      Bruteforce generator, customized with minimum and maximum lengths and an
      alphabet. A dummy process execution function is also available.
      
    - Data-related functions
    
      This encompasses argparse argument types and data transformation functions
      i.e. from and to bin/int/hex.
      
    - Useful decorators
    
      I.e. for trying something while choosing a different action if the
      execution fails (e.g. simply passing, warning the user or interrupting the
      program).
      
    - Input functions
      
      This relates to functions expecting user inputs like a 'pause' function,
      a 'confirm' function (with a customizable message and set of answers), a
      'user_input' function with styling, ...
      
    - Path-related helpers
      
      Some useful path functions, enhanced from pathlib.Path's one. It adds
      various methods to Path and provides helpers for mirroring a path or
      creating a temporary one.
      
    - Timeout items
      
      This provides a timeout decorator and a timeout context manager.
      
    - Others
    
      Namely license-related functions, terminal size get function, ...
""")
for h in __helpers__:
    setattr(ts, h, globals()[h])

__all__ = __features__ = ["pprint", "ts"] + _classprop + _compat + _constants



# -*- coding: UTF-8 -*-
"""Timeout-related functions.

"""
import signal
from contextlib import contextmanager


__all__ = __features__ = ["timeout", "Timeout", "TimeoutError"]


class TimeoutError(Exception):
    pass  # TimeoutError is not handled in Python 2


class Timeout(object):
    """ Timeout context manager.
    
        :param seconds: number of seconds before raising the timeout
        :param message: custom message to be displayed
        :param stop:    whether the execution must be stopped in case of timeout
    """
    def __init__(self, seconds=10, message=None, stop=False):
        self.message = message
        self.seconds = seconds
        self.stop = stop

    def __enter__(self):
        signal.signal(signal.SIGALRM, self._handler)
        signal.alarm(self.seconds)
        return self
    
    def __exit__(self, exc_type, exc_value, exc_traceback):
        signal.signal(signal.SIGALRM, signal.SIG_IGN)
        return not self.stop
    
    def _handler(self, signum, frame):
        raise TimeoutError(self.message or "Execution timeout")


def timeout(seconds=10, message=None, stop=False):
    """ Decorator for applying the Timeout context manager to a function. """
    def _wrapper1(f):
        def _wrapper2(*a, **kw):
            with Timeout(seconds, message, stop) as to:
                return f(*a, **kw)
        return _wrapper2
    return _wrapper1




# -*- coding: UTF-8 -*-
"""Common data transformation functions.

"""
import binascii
from functools import wraps
from math import ceil

from ..types import is_bin, is_bytes, is_hex, is_int, is_list, is_pos_int, is_str
from ..utils import BitArray as Bits
from ...compat import b, ensure_str


__all__ = __features__ = [
    "bin2bin", "bin2int", "bin2hex", "bin2str",
    "hex2bin", "hex2int", "hex2str",
    "int2bin", "int2hex", "int2str", "int2uni",
    "lst2str",
    "str2bin", "str2int", "str2hex", "str2lst",
]


def __validation(**kwargs):
    """ Private generic validation function for the whole data formats. """
    for k, v in kwargs.items():
        if k == "b":
            if not is_bin(v):
                raise ValueError("Bad input binary string '{}'".format(v))
        elif k == "h":
            if not is_hex(v):
                raise ValueError("Bad input hexadecimal string '{}'".format(v))
        elif k == "i":
            if not is_int(v):
                raise ValueError("Bad input integer {}".format(v))
        elif k == "l":
            if not is_list(v):
                raise ValueError("Bad input list {}".format(v))
        elif k in ["n_b", "n_B", "n_c", "n_g"]:
            if not is_pos_int(v, zero=False):
                raise ValueError("Number of {} must be a positive int, not {}"
                                 .format({"n_b": "bits", "n_B": "bits", "n_c": "characters", "n_g": "groups"}[k], v))
        elif k == "o":
            if v not in ["little", "big"]:
                raise ValueError("Bad bits group order '{}'".format(v))
        elif k == "s":
            if not is_str(v):
                raise ValueError("Bad input string of 8-bits characaters '{}'".format(v))
        elif k == "u":
            if not isinstance(v, bool):
                raise ValueError("Bad value for input boolean {}".format(v))
        else:
            if v is None:
                continue
            if not is_pos_int(v, zero=False):
                raise ValueError("{} must be a positive int, not {}".format(k, v))


# BINARY STRING ==> *
def __ensure_bitstring(binary):
    """ Ensure that an input valid binary data is converted to a bitstring. A valid binary data can be a bitstring, a
         list of integer ones/zeros or a list of string ones/zeros. """
    if is_list(binary):
        binary = lst2str(binary, "")
    return binary


def bin2bin(binary_string, nbits_in=8, nbits_out=8):
    """ Convert a binary string with groups of nbits_in bits to a binary string with groups of nbits_out bits. """
    bs = __ensure_bitstring(binary_string)
    __validation(b=bs, n_b=nbits_in, n_B=nbits_out)
    bs = Bits(bs, nbits=nbits_in)
    bs.nbits = nbits_out
    return bs.bin


def bin2hex(binary_string, nbits_in=8, nbits_out=8):
    """ Convert a binary string (eventually using a separator) to a hexadecimal string, using a given number of bits and
         in little or big endian. """
    bs = __ensure_bitstring(binary_string)
    return Bits(bin2bin(bs, nbits_in, nbits_out)).hex


def bin2int(binary_string, nbits_in=8, nbits_out=8, order="big", unsigned=True):
    """ Convert a binary string (eventually using a separator) to an integer, using a given number of bits and in little
         or big endian. """
    bs = __ensure_bitstring(binary_string)
    __validation(b=bs, o=order, u=unsigned)
    bs = Bits(bin2bin(bs, nbits_in, nbits_out))
    pref = ["", "u"][unsigned]
    return getattr(bs, pref + ("intle" if order == "little" else "intbe"))


def bin2str(binary_string, nbits_in=8, nbits_out=8):
    """ Convert a binary string to string of 8-bits characters, using a given number of bits. """
    bs = __ensure_bitstring(binary_string)
    __validation(b=bs, n_b=nbits_in, n_B=nbits_out)
    bs = Bits(bs, nbits=nbits_in)
    bs.nbits = nbits_out
    return bs.bytes


# HEXADECIMAL STRING ==> *
def hex2bin(hex_string, nbits_in=8, nbits_out=8):
    """ Convert a hexadecimal string to a binary string. """
    h = hex_string
    __validation(h=h, n_b=nbits_in, n_B=nbits_out)
    bs = Bits()
    bs.hex = h
    bs._nbits = nbits_in
    bs.nbits = nbits_out
    return bs.bin


def hex2int(hex_string, order="big", unsigned=True):
    """ Convert a hexadecimal string to a big integer. """
    h = hex_string
    __validation(h=h, o=order, u=unsigned)
    bs = Bits()
    bs.hex = h
    pref = ["", "u"][unsigned]
    return getattr(bs, pref + ("intle" if order == "little" else "intbe"))


def hex2str(hex_string):
    """ Convert a hexadecimal string to a string of 8-bits characters. """
    h = hex_string
    __validation(h=h)
    return binascii.unhexlify(b(hex_string))


# INTEGER ==> *
def int2bin(integer, nbits_in=8, nbits_out=8, order="big", unsigned=True):
    """ Convert an integer to a binary string in little or big endian. """
    i = integer
    __validation(i=i, n_b=nbits_in, n_B=nbits_out, o=order, u=unsigned)
    bs = Bits()
    nbits = i.bit_length() + int(not unsigned)
    bs.hex = int(ceil(max(nbits, 1) / 8.0)) * 2 * "0"
    pref = ["", "u"][unsigned]
    setattr(bs, pref + ("intle" if order == "little" else "intbe"), i)
    bs._nbits = nbits_in
    bs.nbits = nbits_out
    return bs.bin


def int2hex(integer, order="big", unsigned=True):
    """ Convert an integer to a hexadecimal string. """
    i = integer
    __validation(i=i, o=order, u=unsigned)
    bs = Bits()
    nbits = i.bit_length() + int(not unsigned)
    bs.hex = int(ceil(max(nbits, 1) / 8.0)) * 2 * "0"
    pref = ["", "u"][unsigned]
    setattr(bs, pref + ("intle" if order == "little" else "intbe"), i)
    return bs.hex


def int2str(integer, order="big", unsigned=True):
    """ Convert a big integer to a string of 8-bits characters. """
    i = integer
    __validation(i=i, o=order, u=unsigned)
    bs = Bits()
    nbits = i.bit_length() + int(not unsigned)
    bs.hex = int(ceil(max(nbits, 1) / 8.0)) * 2 * "0"
    pref = ["", "u"][unsigned]
    setattr(bs, pref + ("intle" if order == "little" else "intbe"), i)
    return bs.bytes


def int2uni(integer):
    """ Convert an integer to a unicode string. """
    i = integer
    __validation(i=i)
    return b("\\u{:0>4}".format(hex(i)[2:])).decode('unicode-escape')


# LIST OF ITEMS ==> STRING
def lst2str(lst, sep=","):
    """ Convert a list of items to a string. """
    __validation(l=lst)
    return sep.join(map(str, lst))


# 8-BITS CHARACTERS ==> *
def str2bin(chars_string, nbits_in=8, nbits_out=8):
    """ Convert a string of 8-bits characters to a binary string. """
    s = chars_string
    __validation(s=s, n_b=nbits_in, n_B=nbits_out)
    bs = Bits(nbits=nbits_in)
    bs.bytes = b(s)
    bs.nbits = nbits_out
    return bs.bin


def str2hex(chars_string):
    """ Convert a string of 8-bits characters to a hexadecimal string. """
    s = chars_string
    __validation(s=s)
    return binascii.hexlify(b(s))


def str2int(chars_string, order="big", unsigned=True):
    """ Convert a string of 8-bits characters to a big integer or, if using blocks of nchars characters, a list of big
         integers. """
    s = chars_string
    __validation(s=s, o=order, u=unsigned)
    bs = Bits()
    bs.bytes = b(s)
    pref = ["", "u"][unsigned]
    return getattr(bs, pref + ("intle" if order == "little" else "intbe"))


def str2lst(chars_string):
    """ Convert a string of 8-bits characters to a list of items, converted into integers if relevant. """
    s = chars_string
    __validation(s=s)
    return [int(c) if is_str(c) and c.isdigit() else c for c in s]


# make conversion functions compatible with input/output strings/bytes
def __fix_inout_formats(f):
    @wraps(f)
    def _wrapper(*args, **kwargs):
        a0 = args[0]
        a0 = ensure_str(a0) if is_str(a0) or is_bytes(a0) else a0
        r = f(a0, *args[1:], **kwargs)
        return r if is_int(r) else b(r) if is_bytes(args[0]) else ensure_str(r)
    return _wrapper


for f in __features__:
    if f in ["int2uni", "str2lst"]:
        continue
    globals()[f] = __fix_inout_formats(globals()[f])


# add multi-input conversion functions
def __items2something(f):
    """ Decorate infmt2outfmt functions to make intfmtS2outfmt (taking multiple inputs, producing a single output. """
    infmt, outfmt = f.__name__.split("2")
    @wraps(f)
    def _wrapper(*args, **kwargs):
        r = [] if outfmt == "int" else ""
        for arg in args:
            out = f(arg, **kwargs)
            if outfmt == "int":
                r.append(out)
            else:
                r += out
        return str2int(ints2str(*r)) if outfmt == "int" else r
    _wrapper.__name__ = "{}s2{}".format(infmt, outfmt)
    _wrapper.__doc__ = f.__doc__
    return _wrapper


def __something2items(f):
    """ Decorate infmt2outfmt functions to make intfmt2outfmtS (taking a single input, producing multiple outputs. """
    infmt, outfmt = f.__name__.split("2")
    @wraps(f)
    def _wrapper(data, **kwargs):
        func = f
        n = kwargs.pop('n_chunks', None)
        inl = kwargs.pop('len_in', None)
        outl = kwargs.pop('len_out', None)
        __validation(n_chunks=n, len_in=inl, len_out=outl)
        if not n and not inl and not outl:
            n = 1
        r = []
        # if input chunk length is given, process the data per block
        if n is None and inl is not None:
            if infmt == "int":
                data = int2bin(data, **kwargs)
                func = globals()['bin2' + outfmt]
            for i in range(0, len(data), inl):
                r.append(func(data[i:i+inl], **kwargs))
        # otherwise, chunk the output
        else:
            if outfmt == "int":
                if n is not None:
                    inl = int(ceil(len(data) / float(n)))
                for i in range(0, len(data), inl):
                    r.append(func(data[i:i+inl], **kwargs))
            else:
                out = func(data, **kwargs)
                if n is not None:
                    outl = int(ceil(len(out) / float(n)))
                for i in range(0, len(out), outl):
                    r.append(out[i:i+outl])
        return r
    _wrapper.__name__ = "{}2{}s".format(infmt, outfmt)
    _wrapper.__doc__ = f.__doc__
    return _wrapper


for fname in __features__[:]:
    f1 = __items2something(globals()[fname])
    globals()[f1.__name__] = f1
    f2 = __something2items(globals()[fname])
    globals()[f2.__name__] = f2
    __features__ += [f1.__name__, f2.__name__]


# FLAGS <=> INTEGER
__features__ += ["flags2int", "int2flags"]


def flags2int(*flags):
    """ Convert a list of booleans to an integer representing binary flags. """
    return int("".join(map(lambda x: "01"[bool(x)], flags)), 2)


def int2flags(integer):
    """ Convert an integer representing binary flags to a list of booleans. """
    i = integer
    __validation(i=i)
    return list(map(lambda x: x == "1", bin(i)[2:]))




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""Tinyscript package information.

"""
import os

__author__    = "Alexandre D'Hondt"
__copyright__ = "© 2017-2020 A. D'Hondt"
__license__   = "AGPLv3 (http://www.gnu.org/licenses/agpl.html)"

with open(os.path.join(os.path.dirname(__file__), "VERSION.txt")) as f:
    __version__ = f.read().strip()



# -*- coding: UTF-8 -*-
"""Common utility functions.

"""
from itertools import cycle, permutations, product
from plyer import notification
from string import printable, punctuation

from .compat import b
from .constants import PYTHON3


__all__ = __features__ = ["bruteforce", "bruteforce_mask", "notify", "strings", "strings_from_file", "xor", "xor_file"]


MASKS = {
    'a': printable,
    'b': "".join(chr(i) for i in range(256)),
    'd': "0123456789",
    'h': "0123456789abcdef",
    'H': "0123456789ABCDEF",
    'l': "abcdefghijklmnopqrstuvwxyz",
    's': " " + punctuation,
    'u': "ABCDEFGHIJKLMNOPQRSTUVWXYZ",
}


def bruteforce(maxlen, alphabet=tuple(map(chr, range(256))), minlen=1, repeat=True):
    """
    Generator for bruteforcing according to minimum and maximum lengths and an alphabet.
    
    :param maxlen:   maximum bruteforce entry length
    :param alphabet: bruteforce alphabet to be used
    :param minlen:   minimum bruteforce entry length (optional)
    :param repeat:   whether alphabet characters can be repeated or not
    :yield:          bruteforce entry
    """
    for i in range(minlen, maxlen + 1):
        if repeat:
            for c in product(alphabet, repeat=i):
                yield c if isinstance(c[0], int) else ''.join(c)
        else:
            for c in permutations(alphabet, i):
                yield c if isinstance(c[0], int) else ''.join(c)


def bruteforce_mask(mask, charsets=None):
    """
    Generator for bruteforcing according to a given mask (similar to this used in hashcat).
     
    :param mask:     bruteforce mask
    :param charsets: custom alphabets for use with the mask
    """
    iterables, charset = [], False
    masks = {k: v for k, v in MASKS.items()}
    masks.update(charsets or {})
    for c in mask:
        if c == "?":
            charset = True
            continue
        if charset:
            iterables.append(masks[c])
            charset = False
            continue
        iterables.append(c)
    for c in product(*iterables):
        yield c if isinstance(c[0], int) else ''.join(c)


def notify(title="", message="", app="", icon="", timeout=5, ticker=""):
    """
    Shortcut to plyer.notification.notify, not considering the 'toast' option, and fail-safe.
    
    :param title:   title of the notification
    :param message: message of the notification
    :param app:     name of the app launching this notification
    :param icon:    name or path of the icon to be displayed along with the message
    :param timeout: time to display the message for
    :param ticker:  text to display on status bar as the notification arrives
    """
    try:
        notification.notify(title, message, app, icon, timeout, ticker)
    except NotImplementedError:
        pass


def strings(data, minlen=4, alphabet=printable):
    """
    Generator yielding strings according to a charset and a minimal length from a given string buffer.

    :param data:     input data
    :param minlen:   minimal length of strings to be considered
    :param alphabet: valid charset for the strings
    """
    result = ""
    for c in b(data):
        if c in b(alphabet):
            result += chr(c) if PYTHON3 else c
            continue
        if len(result) >= minlen:
            yield result
        result = ""
    if len(result) >= minlen:
        yield result


def strings_from_file(filename, minlen=4, alphabet=printable, offset=0):
    """
    Generator yielding strings according to a charset and a minimal length from a given file.
    
    :param filename: input file
    :param minlen:   minimal length of strings to be considered
    :param alphabet: valid charset for the strings
    :param offset:   start offset in the input file
    """
    with open(filename, 'rb') as f:
        f.seek(offset)
        result = ""
        while True:
            data = f.read(max(1024, 2 * minlen))
            if not data:
                break
            for c in data:
                if c in b(alphabet):
                    result += chr(c) if PYTHON3 else c
                    continue
                if len(result) >= minlen:
                    yield result
                result = ""
        if len(result) >= minlen:
            yield result


def xor(str1, str2, offset=0):
    """
    Function for XORing two strings of different length. Either the first or the second string can be longer.

    :param str1:   first string, with length L1
    :param str2:   second string, with length L2
    :param offset: ASCII offset to be applied on each resulting character
    """
    convert = isinstance(str1[0], int) or isinstance(str2[0], int)
    r = b("") if convert else ""
    for c1, c2 in zip(cycle(str1) if len(str1) < len(str2) else str1, cycle(str2) if len(str2) < len(str1) else str2):
        c1 = c1 if isinstance(c1, int) else ord(c1)
        c2 = c2 if isinstance(c2, int) else ord(c2)
        c = chr(((c1 ^ c2) + offset) % 256)
        r += b(c) if convert else c
    return r


def xor_file(filename, key, offset=0):
    """
    Function for XORing a file with a given key.

    :param filename: input file
    :param key:      XOR key
    :param offset:   start offset in the input file
    """
    with open(filename, 'rb+') as f:
        cursor, l = offset, len(key)
        f.seek(cursor)
        while True:
            data = f.read(l)
            if not data:
                break
            f.seek(cursor)
            f.write(xor(data, b(key[:len(data)])))
            cursor += l

